package de.srsoftware.belegscanner;

import de.srsoftware.belegscanner.gui.DocTable;
import de.srsoftware.belegscanner.gui.ImproveSelector;
import de.srsoftware.belegscanner.gui.MainFrame;
import de.srsoftware.belegscanner.gui.ResolutionSelector;
import de.srsoftware.belegscanner.gui.RotationSelector;
import de.srsoftware.belegscanner.gui.StatusBar;
import de.srsoftware.belegscanner.gui.Toolbar;
import de.srsoftware.belegscanner.gui.TypeSelector;
import de.srsoftware.belegscanner.model.api.Page;
import de.srsoftware.belegscanner.model.api.Project;
import de.srsoftware.tools.gui.DateChooser;
import de.srsoftware.tools.gui.SelectComboBox;
import org.json.JSONArray;
import org.json.JSONObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.swing.*;
import javax.swing.text.html.Option;
import java.awt.*;
import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.util.*;
import java.util.List;
import java.util.stream.Stream;

import static de.srsoftware.belegscanner.Application.t;
import static de.srsoftware.belegscanner.Constants.JPG;
import static de.srsoftware.belegscanner.Constants.OCR;
import static de.srsoftware.belegscanner.Constants.PDF;

public class Worker {
    private static final Logger LOG = LoggerFactory.getLogger(Worker.class);
    private static final String HOME = System.getProperty("user.home");
    private static final String PAGE = t(Constants.PAGE);
    private final Configuration config;
    private SelectComboBox pathPicker;
    private StatusBar statusBar;
    private DateChooser datePicker;
    private String profile,rawPath, resolvedPath, type=OCR;
    private Toolbar toolbar;
    private DocTable docTable;
    private boolean improveBrightness,scanning,removePageOnStitching;
    private int resolution,rotationAngle;
    private Dimension dimension;
    private String mode = "Color";
    private MainFrame mainFrame;

    private JTextField heightField, widthField;
    private ResolutionSelector resolutionSelector;
    private ImproveSelector improveSelector;
    private RotationSelector rotationSelector;
    private TypeSelector typeSelector;
    private Map<String,Set<String>> samples = new HashMap<>();


    public Worker(Configuration config) {
        this.config = config;
    }

    public void dropProfile(String droppedProfile) {
        LOG.debug("Worker.dropProfile({})",droppedProfile);
        getProfiles().remove(droppedProfile);
        try {
            config.save();
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }
    public JSONObject getProfiles() {
        return config.getOrCreate("profiles",new JSONObject());
    }

    public Set<String> getSamples(String varName) {
        var set = samples.get(varName);
        if (set == null) set = new HashSet<>();
        LOG.debug("Worker.getSamples(\"{}\") → {}",varName,set);
        return set;
    }

    public String getType() {
        return type;
    }

    public void resolvePath(){
        if (rawPath == null) return;
        LOG.debug("resolve({})",rawPath);
        var date = datePicker.getSelectedDate();
        int year = 1900+date.getYear();
        int month = date.getMonth()+1;
        int day = date.getDate();
        toolbar.resetAdditionalFields();
        resolvedPath = rawPath;
        var last = "";
        var iter = 0;
        do {
            last = resolvedPath;
            LOG.debug("Starting iteration {}: {}",++iter,resolvedPath);
            resolvedPath = resolvedPath.replace("$" + t("HOME"), HOME)
                    .replace("$" + t("PROFILE"), profile)
                    .replace("$" + t("YEAR"), year + "")
                    .replace("$" + t("MONTH"), month < 10 ? "0" + month : "" + month)
                    .replace("$" + t("Month"), monthName(month))
                    .replace("$" + t("month"), monthName(month).toLowerCase())
                    .replace("$" + t("Mon"), monthNameShort(month))
                    .replace("$" + t("mon"), monthNameShort(month).toLowerCase())
                    .replace("$" + t("DAY"), day < 10 ? "0" + day : "" + day);

            var start = resolvedPath.indexOf("$");
            if (start > -1) {
                var end = endOfVar(resolvedPath, start + 1);
                if (end > start + 1) {
                    var variable = resolvedPath.substring(start + 1, end);
                    LOG.debug("…unresolved '{}'", variable);
                    var value = PAGE.equals(variable) ? "[" + variable + "]" : toolbar.requireField(variable);
                    variable = "$" + variable;
                    LOG.debug("replacing '{}' in '{}' by '{}'", variable, resolvedPath, value);
                    resolvedPath = resolvedPath.replace(variable, value);
                    start = resolvedPath.indexOf("$", start + 1);
                } else {
                    start = resolvedPath.indexOf("$", start + 2);
                }
            }
            LOG.debug("Iteration {} ended: {}",iter,resolvedPath);
        } while (!last.equals(resolvedPath));
        LOG.debug("result: {}",resolvedPath);
        toolbar.dropUnusedAdditionalFields();
        statusBar.setPath(resolvedPath+"."+extension());
        mainFrame.pack();
    }

    private String monthName(int month) {
        return switch (month){
            case 1: yield t("January");
            case 2: yield t("February");
            case 3: yield t("March");
            case 4: yield t("April");
            case 5: yield t("May");
            case 6: yield t("June");
            case 7: yield t("July");
            case 8: yield t("August");
            case 9: yield t("September");
            case 10: yield t("October");
            case 11: yield t("November");
            case 12: yield t("December");
            default: yield t("Error");
        };
    }

    private String monthNameShort(int month) {
        return switch (month){
            case 1: yield t("Jan");
            case 2: yield t("Feb");
            case 3: yield t("Mar");
            case 4: yield t("Apr");
            case 5: yield t("May");
            case 6: yield t("Jun");
            case 7: yield t("Jul");
            case 8: yield t("Aug");
            case 9: yield t("Sep");
            case 10: yield t("Oct");
            case 11: yield t("Nov");
            case 12: yield t("Dec");
            default: yield t("Error");
        };
    }

    private String extension() {
        switch (getType()){
            case PDF:
            case OCR:
                return "pdf";
            case JPG:
                return "jpg";
        }
        return "";
    }

    private int endOfVar(String path, int index) {
        while(index<path.length()){
            if (!Character.isAlphabetic(path.codePointAt(index))) return index;
            index++;
        }
        return index;
    }

    public void scan() {
        saveProfile();
        new Thread(this::performScan).start();
    }

    private void saveProfile() {
        var prefix = String.join(".","profiles",profile,"");
        addSamples();
        Map.of(
                "path",rawPath,
                "size",Map.of("height",dimension.height,"width",dimension.width),
                "resolution",resolution,
                "improvements", Map.of("brightness",improveBrightness,"drop_page",removePageOnStitching),
                "rotation",rotationAngle,
                "samples", samples,
                "target_type",type)
                .entrySet()
                .forEach(e -> config.set(prefix+e.getKey(),e.getValue()));
        try {
            config.save();
        } catch (IOException e) {
            LOG.error("Failed to save profile",e);
        }
    }

    private void addSamples() {
        toolbar.additionalFields().forEach(entry -> addSample(entry.getKey(), entry.getValue()));
    }

    private void addSample(String key, String value) {
        var set = samples.get(key);
        if (set == null) samples.put(key, set = new HashSet<>());
        set.add(value);
    }

    private void performScan(){
        LOG.debug("scan()");
        var project = docTable.getProject(resolvedPath).orElseGet(() -> docTable.createProject(resolvedPath));
        int rotation = this.rotationAngle;
        File targetFile = prepareProject(project);
        File jpeg = new File(targetFile+".jpg");
        File pdf = new File(targetFile+".pdf");
        if (!JPG.equals(type) && overwriteAbort(pdf)) return; // if PDF conversion is sheduled: avoid overwriting pdf
        if (overwriteAbort(jpeg)) return;
        var type = getType();
        var page = project.addPage(jpeg);
        docTable.enableButtons(project,false);
        scanJpeg(targetFile,project);
        improveJpeg(targetFile,project,rotation);
        preview(page);

        if (JPG.equals(type)) {
            docTable.setState(project, t("Image scanned"));
            docTable.enableButtons(project,true);
            return;
        };

        convertToPdf(targetFile,type, project);

        if (PDF.equals(type)) {
            page.updateFile(new File(targetFile+".pdf"));
            preview(page);
            docTable.setState(project, t("PDF created"));
            docTable.enableButtons(project,true);
            return;
        }

        performTextRecognition(targetFile,project);

        page.updateFile(new File(targetFile+".pdf"));
        docTable.enableButtons(project,true);
        preview(page);
    }

    private boolean overwriteAbort(File targetFile) {
        if (!targetFile.exists()) return false;
        var answer = JOptionPane.showConfirmDialog(mainFrame,t("File \"{}\" does already exist and would be overwritten. Continue?",targetFile),t("Overwrite file"),JOptionPane.YES_NO_OPTION);
        return answer != JOptionPane.YES_OPTION;
    }

    private void performTextRecognition(File targetFile, Project project) {
        String pdf = targetFile.getName() + ".tmp.pdf";
        String ocr = targetFile.getName() + ".pdf";
        File folder = targetFile.getParentFile();
        var cmd = new Vector<String>();
        cmd.add("pdfsandwich");
        cmd.add("-lang");
        cmd.add("deu");
        cmd.add("-rgb");
        if (resolution>300) {
            cmd.add("-resolution");
            cmd.add(""+resolution);
        }
        cmd.add("-o");
        cmd.add(ocr);
        cmd.add(pdf);
        LOG.debug("executing {}",cmd);
        var builder = new ProcessBuilder(cmd);
        builder.directory(folder);
        try {
            Process process = builder.start();
            docTable.setState(project,t("Text recognition…"));
            int errorCode = process.waitFor();
            if (errorCode != 0) {
                LOG.error("error code: {} for {}",errorCode,cmd);
                docTable.setState(project, t("OCR failed."));
                return;
            }
            if (!folder.toPath().resolve(ocr).toFile().exists()){
                docTable.setState(project,t("OCR failed."));
                return;
            }
        } catch (InterruptedException | IOException e) {
            LOG.error("{} terminated: ",builder,e);
        }

        LOG.debug("removing {}",pdf);
        folder.toPath().resolve(pdf).toFile().delete();
        docTable.setState(project,t("Finished text recognition."));
    }

    private void convertToPdf(File targetFile, String type, Project project) {
        var extension = PDF.equals(type) ? ".pdf" : ".tmp.pdf";
        var raw = rawFile(targetFile);
        String pdf = targetFile.getName() + extension;
        File folder = targetFile.getParentFile();
        var cmd = new Vector<String>();
        cmd.add("convert");
        cmd.add(raw);
        cmd.add(pdf);
        var builder = new ProcessBuilder(cmd);
        builder.directory(folder);
        try {
            Process process = builder.start();
            docTable.setState(project,t("Converting to PDF…"));
            int errorCode = process.waitFor();
            if (errorCode != 0) {
                LOG.error("Error code: {} for {}",errorCode,cmd);
                docTable.setState(project, t("Conversion failed."));
                return;
            }

        } catch (InterruptedException | IOException e) {
            LOG.error("{} terminated: ",builder,e);
        }
        LOG.debug("removing {}",raw);
        folder.toPath().resolve(raw).toFile().delete();
    }

    private String rawFile(File targetFile) {
        return targetFile.getName()+".jpg";
    }

    private void improveJpeg(File target, Project project, int rotation) {
        var raw = rawFile(target);
        File folder = target.getParentFile();
        if (improveBrightness) {
            Vector<String> cmd = new Vector<>();
            cmd.add("mogrify");
            cmd.add("-contrast-stretch");
            cmd.add("0x60%");
            cmd.add(raw);
            var builder = new ProcessBuilder(cmd);
            builder.directory(folder);
            try {
                Process process = builder.start();
                docTable.setState(project,t("Adjusting brightness…"));
                int errorCode = process.waitFor();
                if (errorCode != 0) {
                    LOG.error(t("Error code: {} for {}"),errorCode,cmd);
                    docTable.setState(project, t("Adjustment failed."));
                    setScanning(false);
                    return;
                }
            } catch (InterruptedException | IOException e) {
                LOG.error(t("{} terminated: "),builder,e);
            }
        }
        if (rotation != 0){
            Vector<String> cmd = new Vector<>();
            cmd.add("mogrify");
            cmd.add("-rotate");
            cmd.add(""+rotation);
            cmd.add(raw);
            var builder = new ProcessBuilder(cmd);
            builder.directory(folder);
            try {
                Process process = builder.start();
                docTable.setState(project,t("Rotating image…"));
                int errorCode = process.waitFor();
                if (errorCode != 0) {
                    LOG.error(t("Error code: {} for {}"),errorCode,cmd);
                    docTable.setState(project, t("Rotation failed."));
                    setScanning(false);
                    return;
                }
            } catch (InterruptedException | IOException e) {
                LOG.error(t("{} terminated: "),builder,e);
            }
        }
    }

    private File prepareProject(Project project) {
        var targetFile = project.nextFile();
        docTable.setState(project,t("Scanning…"));
        LOG.debug("scanning as {}",targetFile);
        File folder = targetFile.getParentFile();
        if (!folder.exists()) {
            LOG.warn(t("Path '{}' does not exist!"),folder);
            folder.mkdirs();
        }
        return targetFile;

    }

    private void scanJpeg(File targetFile, Project project) {
        setScanning(true);
        LOG.debug("scanning as {}",targetFile);
        File folder = targetFile.getParentFile();
        if (!folder.exists()) {
            LOG.warn(t("Path '{}' does not exist!"),folder);
            folder.mkdirs();
        }

        var raw = rawFile(targetFile);

        Vector<String> cmd = new Vector<>();
        cmd.add("scanimage");
        cmd.add("-x");
        cmd.add(dimension.width+"");
        cmd.add("-y");
        cmd.add(dimension.height+"");
        cmd.add("--mode");
        cmd.add(mode);
        cmd.add("--resolution");
        cmd.add(resolution+"");
        cmd.add("-o");
        cmd.add(raw);
        LOG.debug("executing {}",cmd);

        ProcessBuilder builder = new ProcessBuilder(cmd);
        builder.directory(folder);
        try {
            Process process = builder.start();
            docTable.setState(project,t("Scanning…"));
            int errorCode = process.waitFor();
            if (errorCode != 0) {
                LOG.error(t("Error code: {} for {}"),errorCode,cmd);
                docTable.setState(project, t("Scan failed."));
                setScanning(false);
                throw new RuntimeException(t("Scanning failed, error code = {}",errorCode));
            }
        } catch (InterruptedException | IOException e) {
            LOG.error(t("{} terminated: "),builder,e);
        }
        setScanning(false);
    }

    public void preview(Page page) {
        mainFrame.preview(page == null ? null : page.file());
    }

    private void setScanning(boolean isScanning) {
        scanning = isScanning;
        if (scanning) {
            toolbar.disableScan();
        } else {
            toolbar.enableScan();
        }
    }


    public void setDimension(Dimension dimension) {
        LOG.debug("setDimension({})",dimension);
        this.dimension = dimension;
    }

    public void setPath(String path) {
        LOG.debug("setPath({})",path);
        this.rawPath = path;
        resolvePath();

    }
    public SelectComboBox setPathPicker(SelectComboBox select) {
        pathPicker = select;
        return pathPicker.onUpdateText(this::setPath);
    }

    public void setStatusBar(StatusBar statusBar) {
        this.statusBar = statusBar;
    }

    public void setType(String newType) {
        LOG.debug("setType({})",newType);
        type = newType;
        resolvePath();
    }

    public void loadProfile(String newProfile) {
        LOG.debug("setCat({}})",newProfile);
        profile = newProfile;
        var prefix = String.join(".","profiles",profile,"");

        Optional.ofNullable((String)config.get(prefix+"path")).ifPresent(path -> {
            pathPicker.setText(path);
            setPath(path);
        });
        Optional.ofNullable((Integer)config.get(prefix+"size.height")).ifPresent(h -> {
            heightField.setText(""+h);
            dimension = new Dimension(dimension.width,h);
        });
        Optional.ofNullable((Integer)config.get(prefix+"size.width")).ifPresent(w -> {
            widthField.setText(""+w);
            dimension = new Dimension(w,dimension.height);
        });
        Optional.ofNullable((Integer)config.get(prefix+"resolution")).ifPresent(res -> {
            resolutionSelector.set(res);
            resolution = res;
        });
        Optional.ofNullable((Boolean)config.get(prefix+"improvements.brightness")).ifPresent(b -> {
                improveSelector.setImproveBrightness(b);
                improveBrightness = b;
        });
        Optional.ofNullable((Boolean)config.get(prefix+"improvements.drop_page")).ifPresent(d -> {
            improveSelector.setDropPage(d);
            removePageOnStitching = d;
        });
        Optional.ofNullable((Integer)config.get(prefix+"rotation")).ifPresent(angle -> {
            rotationSelector.set(angle);
            rotationAngle = angle;
        });
        Optional.ofNullable(config.get(prefix+"target_type")).map(Object::toString).ifPresent(t -> {
                typeSelector.set(t);
                type = t;
        });
        Optional.ofNullable(config.get(prefix+"samples")).filter(o -> o instanceof JSONObject json).map(JSONObject.class::cast).ifPresent(this::processSamples);
    }

    private void processSamples(JSONObject json) {
        samples.clear();
        for (String key : json.keySet()){
            if (json.get(key) instanceof JSONArray arr){
                var set = new HashSet<String>();
                arr.toList().stream().filter(o -> o instanceof String).map(String.class::cast).forEach(set::add);
                samples.put(key,set);
            }
        }
        LOG.debug("processSamples({}),json → {}",json,samples);
    }


    public void setDatePicker(DateChooser pd) {
        datePicker = pd;
        pd.addActionListener(ev -> this.resolvePath());
    }

    public void setToolbar(Toolbar toolbar) {
        this.toolbar = toolbar;
    }

    public void setDocTable(DocTable docTable) {
        this.docTable = docTable;
    }

    public int setResolution(int resolution) {
        return this.resolution = resolution;
    }

    public void setImproveBrightness(boolean newValue) {
        improveBrightness = newValue;
    }

    public Worker setMainFrame(MainFrame mainFrame) {
        this.mainFrame = mainFrame;
        return this;
    }

    public Worker setRotation(int angle) {
        LOG.debug("setRotation({})",angle);
        rotationAngle = angle;
        return this;
    }

    public void join(Project project) {
        File targetFile = prepareProject(project);

        File folder = targetFile.getParentFile();
        var cmd = new Vector<String>();
        cmd.add("pdftk");
        var pages = project.pages();
        Integer firstPage = null;
        Integer lastPage = null;
        int count = 0;
        var joined = new ArrayList<Page>();
        for (var page : pages){
            count++;
            var name = page.file().getName();
            if (name.toLowerCase().endsWith(".pdf")) {
                if (firstPage == null) firstPage = count;
                lastPage = count;
                cmd.add(name);
                joined.add(page);
            }
        }
        cmd.add("cat");
        cmd.add("output");
        var pageReplace = removePageOnStitching ? "" : firstPage+"…"+lastPage;
        File joinedPDF = new File(project.pattern().replace("["+PAGE+"]",pageReplace).trim()+".pdf");
        cmd.add(joinedPDF.getName());
        if (overwriteAbort(joinedPDF)) return;

        LOG.debug("executing {}",cmd);
        var builder = new ProcessBuilder(cmd);
        builder.directory(folder);
        try {
            Process process = builder.start();
            docTable.setState(project,t("Joining pdfs…"));
            int errorCode = process.waitFor();
            if (errorCode != 0) {
                LOG.error("error code: {} for {}",errorCode,cmd);
                docTable.setState(project, t("OCR failed."));
                return;
            }
        } catch (InterruptedException | IOException e) {
            LOG.error("{} terminated: ",builder,e);
        }
        joined.stream().map(Page::file).forEach(File::delete);
        pages.removeAll(joined);
        docTable.setState(project,t("PDFs joined."));
        preview(project.addPage(joinedPDF));
    }

    public void showHelp() {
        String url = "https://git.srsoftware.de/StephanRichter/BelegScanner/src/branch/main/doc/manual-"+ Locale.getDefault().getLanguage().toLowerCase()+".md";
        try {
            Desktop.getDesktop().browse(new URI(url));
        } catch (Exception e) {
            LOG.error("Failed to show help",e);
        }
    }

    public void setRemovePageOnStitching(boolean newValue) {
        removePageOnStitching = newValue;
    }

    public void setDimensionPicker(JTextField height, JTextField width) {
        heightField = height;
        widthField = width;
    }

    public void setResolutionSelector(ResolutionSelector newVal) {
        resolutionSelector = newVal;
    }

    public void setImproveSelector(ImproveSelector newVal) {
        improveSelector = newVal;
    }

    public void setRotationSelector(RotationSelector newVal) {
        rotationSelector = newVal;
    }

    public void setTypeSelector(TypeSelector newVal) {
        typeSelector = newVal;
    }

    /**
     * gibt eine Pseudo-Liste wieder: diese Methode wird nur zum Erzeugen der
     * SelectComboBox benutzt. Diese wiederum greift nur auf die Methoden
     * isEmpty() und stream() zurück. Diese werden hier also so implementiert,
     * dass sie intern auf die Profil-Liste zugreifen.
     * @return
     */
    public List<String> getProfileNames() {
        return new Vector<>(){
            @Override
            public synchronized boolean isEmpty() {
                return getProfiles().isEmpty();
            }

            @Override
            public Stream<String> stream() {
                return getProfiles().keySet().stream();
            }
        };
    }
}
