package de.srsoftware.belegscanner.gui;

import de.srsoftware.belegscanner.Worker;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.swing.*;
import java.awt.*;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;

import static de.srsoftware.belegscanner.Application.t;
import static de.srsoftware.belegscanner.Constants.DEFAULT_HEIGHT;
import static de.srsoftware.belegscanner.Constants.DEFAULT_WIDTH;
import static java.awt.BorderLayout.NORTH;
import static java.awt.BorderLayout.SOUTH;
import static javax.swing.BoxLayout.X_AXIS;
import static javax.swing.BoxLayout.Y_AXIS;

public class FormatSelector extends JPanel {
    private static final Logger LOG = LoggerFactory.getLogger(FormatSelector.class);
    private final JTextField height, width;
    private final Worker worker;

    public FormatSelector(Worker worker){
        setLayout(new BoxLayout(this,Y_AXIS));
        this.worker = worker;
        JPanel widthSelector = new JPanel();
        widthSelector.setLayout(new BoxLayout(widthSelector,X_AXIS));
        widthSelector.setMaximumSize(new Dimension(600, 40));

        widthSelector.add(new JLabel(t("Width")+": "));
        widthSelector.add(width = new JTextField(DEFAULT_WIDTH));
        widthSelector.add(new JLabel("px"));

        add(widthSelector,NORTH);

        var heightSelector = new JPanel();
        heightSelector.setLayout(new BoxLayout(heightSelector, X_AXIS));
        heightSelector.setMaximumSize(new Dimension(600, 40));

        heightSelector.add(new JLabel(t("Height")+": "));
        heightSelector.add(height = new JTextField(DEFAULT_HEIGHT));
        heightSelector.add(new JLabel("px"));
        add(heightSelector,SOUTH);

        KeyAdapter dimensionListener = new KeyAdapter() {
            @Override
            public void keyReleased(KeyEvent e) {
                LOG.debug("updateDimensions()");
                try {
                    updateDim();
                } catch (NumberFormatException nfe) {
                    LOG.warn("Invalid dimensions!");
                }
            }
        };

        width.addKeyListener(dimensionListener);
        height.addKeyListener(dimensionListener);
        worker.setDimensionPicker(height,width);
        updateDim();
    }

    private void updateDim() {
        int w = Integer.parseInt(width.getText().trim());
        int h = Integer.parseInt(height.getText().trim());
        worker.setDimension(new Dimension(w < 0 ? 0 : w, h < 0 ? 0 : h));
    }
}
